@@ -1,4 +1,6 @@
 import io
+import datetime
+from copy import deepcopy
 from pathlib import Path
 from typing import Optional, Iterable, Literal, Any, Union, List, Tuple, Dict
 
@@ -12,19 +14,26 @@ from pyoptimum import AsyncClient
 
 class Model:
 
-    def __init__(self, data: dict):
-        self.r = np.array(data['r'])
-        self.F = np.array(data['F'])
-        self.Q = np.array(data['Q'])
+    def __init__(self, data: Union[dict, "Model"]):
+        if isinstance(data, Model):
+            # copy constructor
+            for k, v in data.__dict__.items():
+                setattr(self, k, deepcopy(v))
 
-        self._std = None
-        self._Di = None
-        self._D = None
-        if 'Di' in data:
-            assert 'D' not in data, "Di and D cannot be both in model data"
-            self.Di = np.array(data['Di'])
         else:
-            self.D = np.array(data['D'])
+            # from dict
+            self.r = np.array(data['r'])
+            self.F = np.array(data['F'])
+            self.Q = np.array(data['Q'])
+
+            self._std = None
+            self._Di = None
+            self._D = None
+            if 'Di' in data:
+                assert 'D' not in data, "Di and D cannot be both in model data"
+                self.Di = np.array(data['Di'])
+            else:
+                self.D = np.array(data['D'])
 
     @property
     def std(self):
@@ -62,11 +71,19 @@ class Model:
         self._D = None
         self._std = None
 
-    def to_dict(self, fields: Optional[Iterable]=None, as_list: bool=False) -> dict:
+    def to_dict(self, fields: Optional[Iterable]=None,
+                as_list: bool=False,
+                normalize=False) -> dict:
+        # normalize
+        alpha = np.max(self.std) ** 2 if normalize else 1.0
         if fields:
             d = {f: getattr(self, f) for f in fields}
+            if normalize:
+                for f in ['Q', 'D']:
+                    if f in fields:
+                        d[f] /= alpha
         else:
-            d = { 'r': self.r, 'D': self.D, 'F': self.F, 'Q': self.Q, 'std': self.std }
+            d = { 'r': self.r, 'D': self.D / alpha, 'F': self.F, 'Q': self.Q / alpha, 'std': self.std }
         return {k: v.tolist() for k, v in d.items()} if as_list else d
 
 
@@ -76,7 +93,6 @@ GREATER_THAN_OR_EQUAL = "\u2265"
 class Portfolio:
 
     ModelMethodLiteral = Literal['linear', 'linear-fractional']
-    ReturnModelLiteral = Literal['median', 'mean']
 
     MethodLiteral = Literal['approximate', 'optimal', 'none']
     ConstraintFunctionLiteral = Literal['purchases', 'sales', 'holdings', 'short sales']
@@ -133,7 +149,13 @@ class Portfolio:
 
         # get model data
         model = self.get_model()
-        data = model.to_dict(('r', 'Q', 'D', 'F'), as_list=True)
+        data = model.to_dict(('r', 'Q', 'D', 'F'),
+                             as_list=True,
+                             normalize=True)
+
+        # has regularization
+        if rho > 0:
+            data['rho'] = rho
 
         # get portfolio data
         value0 = self.portfolio['value ($)'].sum()
@@ -162,10 +184,6 @@ class Portfolio:
             xup = self.portfolio['upper'] * self.portfolio['close ($)'] / value
             data['xup'] = xup.tolist()
 
-        # has regularization
-        if rho > 0:
-            data['rho'] = rho
-
         return data
 
     @staticmethod
@@ -225,6 +243,24 @@ class Portfolio:
         """
         return bool(self.models)
 
+    def set_models(self, models: Dict[str, Union[dict, Model]],
+                   model_weights: Optional[Dict[str, float]]=None) -> None:
+        """
+        Set portfolio models
+
+        :param models: a dictionary with the models per range
+        :param model_weights: the model weights
+        """
+        # add models
+        self.models = {rg: Model(data) for rg, data in models.items()}
+
+        # set model weights
+        model_weights = model_weights or {rg: 1.0 for rg in models.keys()}
+        self.set_models_weights(model_weights)
+
+        # reinitialize frontier
+        self.invalidate_frontier()
+
     def get_model(self) -> Model:
         """
         :return: the portfolio model for the current model and weights
@@ -307,16 +343,16 @@ class Portfolio:
     async def retrieve_models(self,
                               market_tickers: List[str],
                               ranges: Union[str, List[str]],
-                              return_model: ReturnModelLiteral = 'median',
+                              end: datetime.date = datetime.date.today(),
                               common_factors: bool = False,
                               include_prices: bool = False,
                               model_weights: Dict[str, float] = None) -> None:
         """
-        Retrieve the portfolio models based on market tickers
+        Retrieve portfolio models based on market tickers
 
         :param market_tickers: the market tickers
         :param ranges: the ranges to retrieve the portfolio models
-        :param return_model: the type of return model
+        :param end: the last day to retrieve models
         :param common_factors: whether to keep factors common
         :param include_prices: whether to include prices on results
         :param model_weights: the model weights
@@ -326,28 +362,21 @@ class Portfolio:
         data = {
             'tickers': self.portfolio.index.tolist(),
             'market': market_tickers,
+            'end': str(end),
             'range': ranges,
             'options': {
                 'common': common_factors,
-                'return_model': return_model,
                 'include_prices': include_prices
             }
         }
-        model = await self.model_client.call('model', data)
+        models = await self.model_client.call('model', data)
 
         if include_prices:
             # update prices
-            self._update_prices(model.pop('prices'))
+            self._update_prices(models.pop('prices'))
 
-        # add models
-        self.models = {rg: Model(data) for rg, data in model.items()}
-
-        # set model weights
-        model_weights = model_weights or {rg: 1.0 for rg in model.keys()}
-        self.set_models_weights(model_weights)
-
-        # reinitialize frontier
-        self.invalidate_frontier()
+        # set models
+        self.set_models(models, model_weights)
 
     async def retrieve_frontier(self,
                                 cashflow: float, max_sales: float,
@@ -593,16 +622,17 @@ class Portfolio:
             return
 
         # make sure value is array
-        if isinstance(value, float):
+        if isinstance(value, (int, float)):
             value = [value] * len(tickers)
-        value = np.array(value)
+        value = np.array(value, dtype='float64')
 
         # make sure value is in shares
-        shares = self.portfolio.loc[tickers, 'shares']
+        shares = self.portfolio.loc[tickers, 'shares'].values
         if unit == 'value':
-            value /= self.portfolio.loc[tickers, 'close ($)']
+            close = self.portfolio.loc[tickers, 'close ($)'].values
+            value /= close
         elif unit == 'percent value':
-            value *= shares / self.portfolio.loc[tickers, 'close ($)']
+            value *= shares / 100
 
         # initialize
         lb, ub = None, None
